/*
 * msmtp.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003, 2004
 * Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   msmtp is released under the GPL with the additional exemption that
 *   compiling, linking, and/or using OpenSSL is allowed.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#include "exits.h"
#endif

#include "paths.h"
#include "conffile.h"
#include "merror.h"
#include "list.h"
#include "esmtp.h"
#include "net.h"
#include "tls.h"
#include "xmalloc.h"

/* Default configuration file. */
#ifdef WINDOWS
#define CONFFILE "msmtprc.txt"
#elif defined (DJGPP)
#define CONFFILE "_msmtprc"
#else /* UNIX */
#define CONFFILE ".msmtprc"
#endif

/* The name of this program */
char *prgname;


/*
 * The main function processes the options.
 * It returns values from sysexits.h (like sendmail does).
 */

int main(int argc, char *argv[])
{
    /* the configuration */
    int print_version;
    int print_help;
    int print_conf;
    int debug;
    int execute;
    char *conffile;
    char *account_id;
    list_t *account_list;
    account_t *account;
    /* the recipients of the mail */
    int rcptc;
    char **rcptv;
    /* misc */
    int error_code;
    int c;
#ifdef HAVE_GETOPT_LONG
    int option_index;
    struct option options[] =
    {
	{ "version", no_argument, 0, 'v' },
	{ "help",    no_argument, 0, 'h' },
	{ "pretend", no_argument, 0, 'p' },
	{ "debug",   no_argument, 0, 'd' },
	{ "file",    required_argument, 0, 'f' },
	{ "account", required_argument, 0, 'a' },
	{ 0, 0, 0, 0 }
    };
#endif /* HAVE_GETOPT_LONG */
    
    /* process the command line */
    prgname = get_prgname(argv[0]);
    
    error_code = 0;
    print_version = 0;
    print_help = 0;
    print_conf = 0;
    debug = 0;
    execute = 1;
    conffile = NULL;
    account_id = NULL;
    for (;;)
    {
#ifdef HAVE_GETOPT_LONG
	c = getopt_long(argc, argv, "vhpdf:a:", options, &option_index);
#else
	c = getopt(argc, argv, "vhpdf:a:");
#endif /* HAVE_GETOPT_LONG */
	if (c == -1)
	{
	    break;
	}
	switch(c)
	{
	    case 'v':
		print_version = 1;
		execute = 0;
		break;

	    case 'h':
		print_help = 1;
		execute = 0;
		break;

	    case 'p':
		print_conf = 1;
		execute = 0;
		break;
		
	    case 'd':
		print_conf = 1;
		debug = 1;
		break;
		
	    case 'f':
		conffile = xstrdup(optarg);
		break;

	    case 'a':
		account_id = xstrdup(optarg);
		break;

	    default:
		error_code = 1;
		break;
	}
	if (error_code)
	{
	    break;
	}
    }
    if (error_code)
    {
	return EX_USAGE;
    }

    if (print_version)
    {
	printf("This is %s version %s,\n"
#ifdef HAVE_GNUTLS
		"using GnuTLS for TLS encryption "
#endif /* HAVE_GNUTLS */
#ifdef HAVE_OPENSSL
		"using OpenSSL for TLS encryption "
#endif /* HAVE_OPENSSL */
#ifdef USE_GSASL
		"and GSASL for authentication.\n",
#else
		"and built-in functions for authentication.\n",
#endif /* USE_GSASL */
		PACKAGE_NAME, VERSION);
    }
    if (print_help)
    {
	printf("Usage: %s [option...] [--] recipient...\n"
		"Options are:\n"
#ifdef HAVE_GETOPT_LONG
		"-v, --version        print version and exit\n"
		"-h, --help           print help and exit\n"
		"-p, --pretend        print configuration info and exit\n"
		"-d, --debug          print debugging information\n"
		"-f, --file=filename  use `filename' as configuration file\n"
		"-a, --account=id     use account `id' instead of `default'\n",
#else
		"-v           print version and exit\n"
		"-h           print help and exit\n"
		"-p           print configuration info and exit\n"
		"-d           print debugging information\n"
		"-f filename  use `filename' as configuration file\n"
		"-a id        use account `id' instead of `default'\n",
#endif /* HAVE_GETOPT_LONG */
		prgname);
    }
    
    if (!execute && !print_conf)
    {
	return EX_OK;
    }
    /* get the list of recipients */
    rcptc = argc - optind;
    rcptv = &(argv[optind]);
    if (rcptc == 0 && execute)
    {
	add_error_line("no recipients given");
	print_error_lines(stderr);
	return EX_USAGE;
    }

    /* read the configuration file */
    if (!conffile)
    {
	if (!(conffile = get_default_conffile(CONFFILE)))
	{
	    add_error_line("cannot get default configuration file name");
	    print_error_lines(stderr);
	    return EX_NOINPUT;
	}
    }
    if ((error_code = get_conf(conffile, &account_list)) != 0)
    {
	add_error_line("could not use configuration file `%s'", conffile);
	print_error_lines(stderr);
	free(conffile);
	free(account_id);
	if (error_code == 1)
	{
	    return EX_IOERR;
	}
	else
	{
	    return EX_DATAERR;
	}
    }

    /* get the account data */
    if (!account_id)
    {
	account_id = xstrdup("default");
    }
    if (!(account = find_account(account_list, account_id)))
    {
	add_error_line("no account `%s' found in configuration file `%s'",
		account_id, conffile);
	print_error_lines(stderr);
	free(conffile);
	free(account_id);
	return EX_DATAERR;
    }

    if (print_conf)
    {
    	printf("using account `%s' from `%s':\n"
		"host            = %s\n"
		"port            = %d\n"
		"domain          = %s\n"
		"from            = %s\n",
		account->id, conffile,
		account->host,
		account->port,
		account->domain,
		account->from);
	printf("auth            = ");
	if (!account->auth_mech)
	{
	    printf("none\n");
	}
	else if (account->auth_mech[0] == '\0')
	{
	    printf("choose\n");
	}
	else
	{
	    printf("%s\n", account->auth_mech);
	}
	printf("user            = %s\n"
		"password        = %s\n"
		"tls             = %s\n"
		"tls_trust_file  = %s\n"
		"tls_key_file    = %s\n"
		"tls_cert_file   = %s\n"
		"tls_nostarttls  = %s\n"
		"tls_nocertcheck = %s\n"
		"dsn_notify      = %s\n"
		"dsn_return      = %s\n"
		"keepbcc         = %s\n",
		account->username ? account->username : "",
		account->password ? "*" : "",
		account->tls ? "true" : "false", 
		account->tls_trust_file ? account->tls_trust_file : "",
		account->tls_key_file ? account->tls_key_file : "",
		account->tls_cert_file ? account->tls_cert_file : "",
		account->tls_nostarttls ? "true" : "false",
		account->tls_nocertcheck ? "true" : "false",
		account->dsn_notify ? account->dsn_notify : "",
		account->dsn_return ? account->dsn_return : "",
		account->keepbcc ? "true" : "false");
    }
    if (!execute)
    {
	return EX_OK;
    }
    
    if (net_init() != 0)
    {
	add_error_line("cannot initialize networking library");
	print_error_lines(stderr);
    	free(conffile);
	free(account_id);
	list_xfree(account_list, _account_free);
	return EX_SOFTWARE;	
    }
    if (account->tls)
    {
	if (tls_init() != 0)
	{
	    add_error_line("cannot initialize TLS library");
	    print_error_lines(stderr);
	    free(conffile);
	    free(account_id);
	    list_xfree(account_list, _account_free);
	    return EX_SOFTWARE;
	}
    }
    /* send the mail */
    if ((error_code = smtp_send(
		    account->host, 
		    account->port, 
		    account->domain,
		    account->from,
		    account->auth_mech, 
		    account->username, 
		    account->password, 
		    account->tls, 
		    account->tls_key_file,
		    account->tls_cert_file,
		    account->tls_trust_file,
		    account->tls_nocertcheck, 
		    account->tls_nostarttls, 
		    account->dsn_return,
		    account->dsn_notify,
		    account->keepbcc,
		    rcptc, rcptv, 
		    stdin, 
		    debug)) != EX_OK)
    {
	add_error_line("could not send mail [account: `%s', file: `%s']",
		account_id, conffile);
	print_error_lines(stderr);
    }

    /* clean up */
    if (account->tls)
    {
	tls_deinit();
    }
    net_deinit();
    free(conffile);
    free(account_id);
    list_xfree(account_list, _account_free);

    return error_code;
}
