/*
 * net.c
 * 
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003, 2004
 * Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   msmtp is released under the GPL with the additional exemption that
 *   compiling, linking, and/or using OpenSSL is allowed.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#ifdef WINDOWS

#include <winsock2.h>
#ifdef HAVE_GETADDRINFO
#include <ws2tcpip.h>
#endif /* HAVE_GETADDRINFO */

#elif defined (DJGPP)

#include <unistd.h>
#include <tcp.h>
#include <netdb.h>
#include <errno.h>
extern int errno;

#else /* UNIX */

#include <unistd.h>
#ifndef HAVE_GETADDRINFO
#include <netinet/in.h>
#endif /* !HAVE_GETADDRINFO */
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
extern int errno;

#endif /* UNIX */


#include <string.h>

#include "net.h"
#include "merror.h"
#include "xmalloc.h"


/*
 * net_init()
 *
 * Initializes the networking libraries.
 * Returns 0 on success and 1 on failure.
 */

int net_init(void)
{
#ifdef WINDOWS
   
    WORD wVersionRequested;
    WSADATA wsaData;
    int error_code;
    char *reason;
    
    wVersionRequested = MAKEWORD(2, 0);
    if ((error_code = WSAStartup(wVersionRequested, &wsaData)) != 0)
    {
	switch (error_code)
	{
	    case WSASYSNOTREADY:
		reason = "the underlying network subsystem is not ready";
		break;

	    case WSAVERNOTSUPPORTED:
		reason = "the requested version is not available";
		break;

	    case WSAEINPROGRESS:
		reason = "a blocking operation is in progress";
		break;

	    case WSAEPROCLIM:
		reason = "limit on the number of tasks has been reached";
		break;

	    case WSAEFAULT:
		reason = "invalid request";
		break;

	    default:
		reason = "unknown error";
		break;
	}
	add_error_line("Failed to initialize networking DLL [%s]", reason);
	return 1;
    }
    else
    {
	return 0;
    }
    
#else
    
    return 0;
    
#endif /* WINDOWS */    
}


/*
 * [WINDOWS only] wsa_strerror() 
 *
 * Translates error codes to strings.
 */

#ifdef WINDOWS
char *wsa_strerror(int error_code)
{
    char *error_string;
    
    switch (error_code)
    {
	case WSAENETDOWN:
    	    error_string = "the network subsystem has failed";
	    break;
	    
	case WSAHOST_NOT_FOUND:
    	    error_string = "host not found (authoritative)";
	    break;
	    
	case WSATRY_AGAIN:
    	    error_string = "host not found (nonauthoritative) or server failure";
	    break;
	    
	case WSANO_RECOVERY:
    	    error_string = "nonrecoverable error";
	    break;
	    
	case WSANO_DATA:
    	    error_string = "valid name, but no data record of requested type";
	    break;

	case WSAEAFNOSUPPORT:
	    error_string = "address family not supported";
	    break;
	    
	case WSAEMFILE:
	    error_string = "no socket descriptors available";
	    break;

	case WSAENOBUFS:
	    error_string = "no buffer space available";
	    break;

	case WSAEPROTONOSUPPORT:
	    error_string = "protocol not supported";
	    break;

	case WSAEPROTOTYPE:
	    error_string = "wrong protocol type for this socket";
	    break;

	case WSAESOCKTNOSUPPORT:
	    error_string = "socket type is not supported in this address family";
	    break;
	    
	case WSAEADDRNOTAVAIL:
	    error_string = "remote address is not valid";
	    break;

	case WSAECONNREFUSED:
	    error_string = "connection refused";
	    break;

	case WSAENETUNREACH:
	    error_string = "network unreachable";
	    break;

	case WSAETIMEDOUT:
	    error_string = "timeout";
	    break;
	    
	case WSAENOTCONN:
	    error_string = "socket not connected";
	    break;

	case WSAESHUTDOWN:
	    error_string = "the socket was shut down";
	    break;

	case WSAEHOSTUNREACH:
	    error_string = "host unreachable";
	    break;

	case WSAECONNRESET:
	    error_string = "connection reset by peer";
	    break;
	    
	default:
    	    error_string = "unknown error";
	    break;
    }

    return error_string;
}
#endif /* WINDOWS */


/*
 * close_socket()
 *
 * Closes a socket.
 * [This function is needed because Windows cannot just close() a socket].
 */

int close_socket(int fd)
{
#ifdef WINDOWS
    return closesocket(fd);
#else
    return close(fd);
#endif
}


/*
 * open_socket() 
 *
 * Opens a TCP socket to 'hostname':'port'.
 * 'hostname' may be a host name or an IPv4 address in dot notation.
 * If HAVE_GETADDRINFO is defined, it may also be an an IPv6 address in colon
 * notation.
 * On success, the file descriptor is returned. 
 * On failure, -1 is returned.
 */

int open_socket(char *hostname, int port)
{    
#ifdef HAVE_GETADDRINFO
    
    int fd;
    char *port_string;
    struct addrinfo hints;
    struct addrinfo *res0;
    struct addrinfo *res;
    int error_code;
    int cause;
    
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = 0;
    hints.ai_protocol = 0;
    hints.ai_addrlen = 0;
    hints.ai_canonname = NULL;
    hints.ai_addr = NULL;
    hints.ai_next = NULL;
    xasprintf(&port_string, "%d", port);
    error_code = getaddrinfo(hostname, port_string, &hints, &res0);
    free(port_string);
    if (error_code)
    {
	add_error_line("cannot locate host `%s' [%s]",
		hostname, gai_strerror(error_code));
	return -1;
    }

    fd = -1;
    cause = 0;
    for (res = res0; res; res = res->ai_next)
    {
	fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (fd < 0)
	{
	    cause = 1;
	    continue;
	}
	if (connect(fd, res->ai_addr, res->ai_addrlen) < 0) 
	{
	    cause = 2;
	    close_socket(fd);
	    fd = -1;
	    continue;
	}
	break;
    }
    freeaddrinfo(res0);
    if (fd < 0)
    {
	if (cause == 1)
	{
	    add_error_line("cannot create socket [%s]", 
#ifdef WINDOWS
		    wsa_strerror(WSAGetLastError())
#else
		    strerror(errno)
#endif
		    );
	}
	else if (cause == 2)
	{
	    add_error_line("cannot connect to `%s', port %d [%s]", 
		    hostname, port,
#ifdef WINDOWS
		    wsa_strerror(WSAGetLastError())
#else
		    strerror(errno)
#endif
		    );
	}
	else /* cause == 0, can't happen */
	{
	    add_error_line("cannot locate host `%s' "
		    "[getaddrinfo() returned crap]", hostname);
	}
	return -1;
    }
    
    return fd;

#else /* !HAVE_GETADDRINFO */

    int fd;
    struct sockaddr_in sock;
    struct hostent *remote_host;
    
    if (!(remote_host = gethostbyname(hostname)))
    {
	add_error_line("cannot locate host `%s' [%s]", hostname, 
#ifdef WINDOWS
		wsa_strerror(WSAGetLastError())
#else
		hstrerror(h_errno)
#endif
		);
	return -1;
    }

    if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
	add_error_line("cannot create socket [%s]", 
#ifdef WINDOWS
		wsa_strerror(WSAGetLastError())
#else
		strerror(errno)
#endif
		);
	return -1;
    }
    
    sock.sin_family = AF_INET;
    sock.sin_port = htons((unsigned short int)port);
    memcpy(&sock.sin_addr, remote_host->h_addr_list[0], (size_t)remote_host->h_length);

    if (connect(fd, (struct sockaddr *)(&sock), sizeof(sock)) < 0)
    {
	add_error_line("cannot connect to `%s', port %d [%s]",
		hostname, port,
#ifdef WINDOWS
		wsa_strerror(WSAGetLastError())
#else
		strerror(errno)
#endif
		);
	return -1;
    }

    return fd;

#endif /* HAVE_GETADDRINFO */
}


/*
 * net_getchar()
 *
 * Reads a character from 'fd' and stores it in 'c'.
 * Return value is 0 on success, -1 on EOF and 1 on error.
 */

int net_getchar(int fd, char *c)
{
#ifdef WINDOWS
    int error_code;
    
    if ((error_code = recv(fd, c, 1, 0)) == 1)
    {
	return 0;
    }
    else if (error_code == 0)
    {
	return -1;
    }
    else
    {
	add_error_line("network read error [%s]", wsa_strerror(WSAGetLastError()));
	return 1;
    }
    
#else /* !WINDOWS */
    
    ssize_t error_code;
    
    if ((error_code = read(fd, c, (size_t)1)) == 1)
    {
	return 0;
    }
    else if (error_code == 0)
    {
	return -1;
    }
    else /* error_code == -1 */
    {
	add_error_line("network read error [%s]", strerror(errno));
	return 1;
    }
#endif /* !WINDOWS */
}


/*
 * net_getline()
 * 
 * Reads a line from 'fd' into 'line', which is of size 'size'.
 * The line end (either '\n' or '\r\n') is not stored in 'line'. 
 * 'line' will alway be terminated by '\0'.
 * Return value is 0 on success, 1 on read error and 2 if the line is longer
 * than 'size - 1' characters.
 */

int net_getline(int fd, char *line, int size)
{
    char c;
    int i;
    int error_code;

    i = 0;
    while ((error_code = net_getchar(fd, &c)) == 0)
    {
	if (i == size - 1)
    	{
	    /* line too long */
	    line[i] = '\0';
	    return 2;
	}
	else if (c == '\n')
	{
	    break;
	}
	else
	{
	    line[i++] = c;
	}
    }
    if (c == '\n' && line[i - 1] == '\r')
    {
	line[i - 1] = '\0';
    }
    line[i] = '\0';
    if (error_code == 0 || error_code == -1)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}


/*
 * net_putline()
 *
 * Writes a line to 'fd'.
 * Return value is 0 on success and 1 on write error.
 */

int net_putline(int fd, char *line)
{
#ifdef WINDOWS
    int error_code;
    size_t count;

    count = strlen(line);    
    error_code = send(fd, line, count, 0);
    if (error_code < 0)
    {
	add_error_line("network write error [%s]", wsa_strerror(WSAGetLastError()));
	return 1;
    }
    else if ((size_t)error_code == count)
    {
	return 0;
    }
    else /* 0 <= error_code < count */
    {
	add_error_line("network write error");
	return 1;
    }
    
#else /* !WINDOWS */

    ssize_t error_code;
    size_t count;

    count = strlen(line);    
    error_code = write(fd, line, count);
    if (error_code < 0)
    {
	add_error_line("network write error [%s]", strerror(errno));
	return 1;
    }
    else if ((size_t)error_code == count)
    {
	return 0;
    }
    else /* 0 <= error_code < count */
    {
	add_error_line("network write error");
	return 1;
    }
#endif /* !WINDOWS */
}


/*
 * net_deinit()
 *
 * Deinit networking library
 */

void net_deinit(void)
{
#ifdef WINDOWS
    /* Ignore deinit errors, we cannot do anything about it anyway */
    (void)WSACleanup();
#endif /* WINDOWS */
}
